#!/usr/bin/env python
#
# Copyright (c) 2007, 2008 Agostino Russo
#
# Written by Agostino Russo <agostino.russo@gmail.com>
#
# pack.py is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# pack.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os
import shutil
from modulegraph.find_modules import find_modules
from modulegraph.modulegraph import SourceModule, Package, Script, Extension
from os.path import abspath, join, basename, dirname, exists
import py_compile
from optparse import OptionParser
import logging
from version import *

log = logging.getLogger(__name__)

class Application(object):
    def __init__(self, root_dir):
        self.root_dir = os.path.abspath(root_dir)

    def run(self):
        self.localize()
        self.parse_commandline_arguments()
        self.set_logger()
        self.create_dirs()
        self.add_dependencies(self.main_script)
        self.add_extras()

    def localize(self):
        try:
            import gettext
            gettext.install(APPLICATION_NAME)
        except ImportError:
            import __builtin__
            def dummytrans (text):
                """A _ function for systems without gettext. Effectively a NOOP"""
                return text
            __builtin__.__dict__['_'] = dummytrans

    def print_readme(self):
        readme = ajoin(self.root_dir, 'README')
        readme = open(readme, 'r')
        content = readme.read()
        readme.close()
        print content

    def parse_commandline_arguments(self):
        usage = "%prog [options] main_script [extra extra extra]"
        version = version="%s-rev%s" % (APPLICATION_VERSION, APPLICATION_REVISION)
        description='Pypack packages a script into a directory containing all needed depencies suitable for "folder deployment". Run `python pypack --readme` for more information.'
        parser = OptionParser(usage=usage, description = description, version = version)
        parser.add_option("--readme", action="store_true", dest="readme", help="show more information")
        parser.add_option("--verbose", action="store_true", dest="verbose", help="run in verbose mode, all messages are displayed")
        parser.add_option("--debug", action="store_true", dest="debug", help="print debug messages to stdout")
        parser.add_option("--print", action="store_true", dest="print_dependencies", help="print dependencies and exit")
        parser.add_option("--outputdir", dest="out_dir", default="./build", help="directory where the files will be built, if non existent, it will be created")
        parser.add_option("--bytecompile", action="store_true", dest="bytecompile", help="bytecompile the python files")
        options, args = parser.parse_args()
        self.options, args = parser.parse_args()
        if options.readme:
            print self.print_readme()
            sys.exit(0)
        if not len(args):
            parser.print_help()
            sys.exit(1)
        self.main_script = args[0]
        self.extras = args[1:]
        if self.options.debug:
            self.options.verbose = True
        if self.options.print_dependencies:
            self.print_dependencies()
            sys.exit(0)

    def create_dirs(self):
        self.out_dir = self.options.out_dir
        self.out_dir = ajoin(self.out_dir)
        if os.path.exists(self.out_dir):
            print 'ERROR: the build directory "%s" already exists.\nRemove that before running %s again.' % (self.out_dir, APPLICATION_NAME)
            sys.exit(1)
        self.lib_dir = ajoin(self.out_dir, 'lib')
        os.makedirs(self.lib_dir)

    def set_logger(self):
        '''
        Adjust the application root logger settings
        '''
        root_logger = logging.getLogger()
        if root_logger.handlers:
            handler = root_logger.handlers[0]
        else:
            handler = logging.StreamHandler()
            root_logger.addHandler(handler)
        formatter = logging.Formatter('%(message)s', datefmt='%m-%d %H:%M')
        handler.setFormatter(formatter)
        root_logger.setLevel(logging.DEBUG)
        if self.options.verbose:
            handler.setLevel(logging.DEBUG)
        else:
            handler.setLevel(logging.ERROR)

    def print_dependencies(self):
        scripts = [self.main_script]
        scripts += [f for f in self.extras if f[:-3] == '.py']
        modules = []
        for script in scripts:
            for m in get_modules(script):
                modules.append(m)
        for m in modules:
            print m

    def add_dependencies(self, script):
        '''
        Adds all dependencies to lib_dir
        '''
        modules = get_modules(script)
        lib_dir = self.lib_dir
        for module in modules:
            if not module.filename:
                continue
            if isinstance(module, SourceModule):
                target = module2path(lib_dir, module.identifier)
                target +=  '.py'
                self.compile(module.filename, target)
            elif isinstance(module, Package):
                target = module2path(lib_dir, module.identifier)
                target = ajoin(target, '__init__.py')
                self.compile(module.filename, target)
            elif isinstance(module, Script):
                log.debug("Script %s", script)
                if module.identifier[0] != "_":
                    script_name = basename(script)
                    target = ajoin(self.out_dir, script_name)
                    self.compile(module.filename, target)
            elif isinstance(module, Extension):
                target = module2path(lib_dir, module.identifier)
                target = dirname(target)
                makedirs(target)
                log.debug("copying %s -> %s" % (module.filename, target))
                shutil.copy(module.filename, target)
            else:
                log.error("Unkown module %s", module)

    def compile(self, source, target):
        log.debug("copying %s -> %s", source, target)
        makedirs(dirname(target))
        shutil.copy(source, target)
        if self.options.bytecompile:
            log.debug("compiling %s", target)
            dfile = target
            if target.startswith(self.out_dir):
                dfile = target[len(self.out_dir):]
            py_compile.compile(target, dfile=dfile, doraise=True)
            os.unlink(target)

    def add_extras(self):
        '''
        Add extra files, modules, pacakges or directories
        '''
        for extra in self.extras:
            if os.path.isfile(extra):
                if extra.endswith('.py'):
                    self.add_dependencies(extra)
                    target = os.path.basename(extra)
                    target = ajoin(self.lib_dir, target)
                    self.compile(extra, target)
                else:
                    target = os.path.basename(extra)
                    target = ajoin(self.out_dir, target)
                    log.debug("copying %s -> %s", extra, target)
                    shutil.copy(extra, target)
            elif os.path.isdir(extra):
                target = os.path.basename(extra)
                target = ajoin(self.out_dir, target)
                log.debug("copying %s -> %s", extra, target)
                shutil.copytree(extra, target, symlinks=False)
            else:
                print 'ERROR: The extra argument "%s" cannot be found' % extra
                sys.exit(1)


def ajoin(*args):
    return os.path.abspath(os.path.join(*args))

def makedirs(target_path):
    try:
        os.makedirs(target_path)
    except:
        pass

def module2path(root_dir, module_name):
    path = module_name.split('.')
    path = ajoin(root_dir, *path)
    return path

def get_modules(script_path):
    sys.path.insert(0, os.path.dirname(script_path))
    mf = find_modules((script_path,))
    del sys.path[0]
    modules = mf.flatten()
    return modules



#!/usr/bin/python

# Copyright (C) 2009 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twisted.trial import unittest
#from twisted.internet.defer import Deferred
import dbus
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from usbcreator.backends.udisks import UDisksBackend

import subprocess, os

class TestUDisksBackend(unittest.TestCase):
    def setUp(self):
        fake_udisks = os.path.join(os.path.dirname(__file__),
                                      'fake_udisks.py')
        self.proc = subprocess.Popen(['python', fake_udisks])
        import time
        time.sleep(1)
        #self._startDBus()
        self.bus = dbus.SessionBus()
        self.backend = UDisksBackend(bus=self.bus)
    def tearDown(self):
        self.proc.terminate()
        self.proc.wait()
        #self._stopDBus()

    # XXX evand 2009-06-03: Uninteresting bit of code to test, as we test our
    # fake EnumerateDevices and signal handling elsewhere.

    #def test_detect_devices(self):
    #    d = Deferred()
    #    def handle_reply(res):
    #        print(res)
    #        d.callback(res)
    #    self.backend.detect_devices(cb=handle_reply)
    #    return d

    def test_disk_added(self):
        d = '/org/freedesktop/UDisks/devices/sdb'
        self.backend._device_added(d)

    def test_partition_added(self):
        d = '/org/freedesktop/UDisks/devices/sdb1'
        self.backend._device_added(d)

    def test_cd_added(self):
        d = '/org/freedesktop/UDisks/devices/scd0'
        self.backend._device_added(d)

    # XXX evand 2009-06-03: Just use the 3 second delay for now, then come back
    # to this.  Getting the tests done and the core code written is more
    # important.
    #def _startDBus(self):
    #    """Start our own session bus daemon for testing."""
    #    config_file = os.path.join(os.path.dirname(os.getcwd()), "tests",
    #                               "dbus-session.conf")
    #    dbus_args = ["--fork",
    #                 "--config-file=" + config_file,
    #                 "--print-address"]
    #    p = subprocess.Popen(['/bin/dbus-daemon'] + dbus_args,
    #                         bufsize=4096, stdout=subprocess.PIPE,
    #                         universal_newlines=True)
    #    data = p.stdout

    #    self.dbus_address = "".join(data.readlines()).strip()
    #    self.dbus_pid = p.pid + 1

    #    if self.dbus_address != "":
    #        os.environ["DBUS_SESSION_BUS_ADDRESS"] = self.dbus_address
    #    else:
    #        os.kill(self.dbus_pid, signal.SIGKILL)
    #        raise DBusLaunchError("There was a problem launching dbus-daemon.")

    #def _stopDBus(self):
    #    """Stop our DBus session bus daemon."""
    #    del os.environ["DBUS_SESSION_BUS_ADDRESS"]
    #    os.kill(self.dbus_pid, signal.SIGKILL)




# Copyright (C) 2009 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from usbcreator.frontends.winui.winui import ui

# FIXME evand 2009-07-24: Need to handle the user hitting the close button and
# making the window visible in the pager.
class InstallWindow(ui.MainWindow):
    def __init__(self, main_window):
        self.width = 450
        ui.MainWindow.__init__(self, main_window, 0, 0, self.width, 115,
                               _('Installing'))

    def on_init(self):
        right_edge = self.width - 30
        self.progress_title = ui.Label(self, 10, 5, right_edge, 20,
                                       _('Starting up...'))
        self.progress_bar = ui.ProgressBar(self, 10, 25, right_edge, 25)
        self.progress_subtitle = ui.Label(self, 10, 60, 350, 20, '')
        self.cancel_button = ui.Button(self, right_edge - 50, 60, 60, 20,
                                       _('&Cancel'))



# remtest.py - estimate remaining time of some task
# Copyright (C) 2009  Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time

class RemainingTimeEstimator:

    """Estimate the remaining time of some task.
    
    The estimate is based on measurements of the progress of the task at
    various intervals. The caller occasionally feeds this class how much of a
    task has been achieved, and the total size of the task. The caller can

    The remaining time is estimated based on timings taken from calls to
    the estimate method.
    
    The caller may set the min_age and max_age attributes (defaults are
    one and five seconds, respectively). min_age is the minimum age of
    the measurements before speed is computed. max_age is the maximum age
    of measurements before they are discarded.
    
    """
    
    # The caller provides data via the 'estimate' method. We take the amount
    # of work done and total from the caller, and the current time from
    # time.time. Then we do an estimate by measuring the total amount of
    # work done during the last X seconds, and compute the speed from
    # that. Using the speed, we can compute the remaining time.
    
    def __init__(self):
        self.measurements = []
        self.min_age = 1.0 # seconds
        self.max_age = 5.0 # seconds
        
    def measure(self, done, total, current_time):
        self.measurements.append((done, current_time))
        while self.measurements:
            x, t = self.measurements[0]
            if t >= current_time - self.max_age:
                break
            del self.measurements[0]
        
    def compute_speed(self):
        """Compute speed from current set of measurements."""
        if not self.measurements:
            return None
        done_oldest, time_oldest = self.measurements[0]
        done_newest, time_newest = self.measurements[-1]
        duration = time_newest - time_oldest
        if duration < self.min_age:
            return None
        done = done_newest - done_oldest
        return done / duration
        
    def estimate(self, done, total):
        """Estimate remaining time, given work done and work total.
        
        Return tuple (remaining_time, speed), where time is in seconds,
        and speed is in units of work per second.
        
        If either can't be estimated, return a tuple of two Nones instead.
        
        Neither done nor total should decrease from call to call.
        
        """

        self.measure(done, total, time.time())
        speed = self.compute_speed()
        if speed:
            return (total - done) / speed, speed
        else:
            return None, None



#!/usr/bin/python

# Copyright (C) 2009 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twisted.trial import unittest
from twisted.internet.defer import Deferred

# TODO evand 2009-06-02: Move the common bits into a unittest.TestCase
# subclass like the Ubuntu One client does?
import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

DISKS_IFACE = 'org.freedesktop.UDisks'
DEVICE_IFACE = 'org.freedesktop.UDisks.Device'
PROPS_IFACE = 'org.freedesktop.DBus.Properties'

class TestDbus(unittest.TestCase):
    def setUp(self):
        # TODO evand 2009-07-01: Move to a private bus.  See the notes in
        # test_udisks_backend.py
        self.bus = dbus.SessionBus()
        self.d = FakeUDisks(self.bus, '/org/freedesktop/UDisks')
        self.udisks = self.bus.get_object(DISKS_IFACE,
                                          '/org/freedesktop/UDisks')

    def tearDown(self):
        # FIXME evand 2009-07-06: There must be a better way...
        for dev in self.d.device_objects:
            dev.bus_name.get_bus().release_name(dev.bus_name.get_name())
            dev.remove_from_connection()
            del dev.bus_name
        self.d.bus_name.get_bus().release_name(self.d.bus_name.get_name())
        self.d.remove_from_connection()
        del self.d.bus_name

    def test_enumerate(self):
        d = Deferred()
        def handle_reply(result):
            d.callback(result)
        def handle_error(err):
            d.errback(err)
        self.udisks.EnumerateDevices(dbus_interface=DISKS_IFACE,
                            reply_handler=handle_reply,
                            error_handler=handle_error)
        return d

    def test_props(self):
        d = Deferred()
        def handle_reply(result):
            d.callback(result)
        def handle_error(err):
            d.errback(err)
        path = '/org/freedesktop/UDisks/devices/sdb'
        device = self.bus.get_object(DISKS_IFACE, path)
        device.Get(path,
                   'device-is-system-internal',
                   dbus_interface=PROPS_IFACE,
                   reply_handler=handle_reply,
                   error_handler=handle_error)
        return d

    def test_added(self):
        '''Fires the DeviceAdded signal.'''
        d = Deferred()
        def handle_reply(result):
            d.callback(result)
        def handle_error(err):
            d.errback(err)
        self.bus.add_signal_receiver(handle_reply,
                             signal_name='DeviceAdded',
                             dbus_interface=DISKS_IFACE,
                             path='/org/freedesktop/UDisks')

        self.udisks.emitDeviceAdded(reply_handler=lambda : None,
                                    error_handler=handle_error)
        return d

    def test_removed(self):
        '''Fires the DeviceRemoved signal.'''
        d = Deferred()
        def handle_reply(result):
            d.callback(result)
        def handle_error(err):
            d.errback(err)
        self.bus.add_signal_receiver(handle_reply,
                             signal_name='DeviceRemoved',
                             dbus_interface=DISKS_IFACE,
                             path='/org/freedesktop/UDisks')

        self.udisks.emitDeviceRemoved(reply_handler=lambda : None,
                                      error_handler=handle_error)
        return d

# XXX evand 2009-07-07: Built according to the udisks specification:
# http://hal.freedesktop.org/docs/udisks/

class FakeUDisks(dbus.service.Object):
    def __init__(self, bus, path):
        self.bus_name = dbus.service.BusName(DISKS_IFACE,
                                             bus)
        self.devices = {'/org/freedesktop/UDisks/devices/sdb' :
                            { 'device-is-system-internal' : False,
                              'device-is-drive' : True,
                              'device-is-optical-disc' : False,
                            },
                        '/org/freedesktop/UDisks/devices/sdb1' :
                            { 'device-is-system-internal' : False,
                              'device-is-drive' : False,
                              'device-is-partition' : True,
                              'partition-label' : 'test-label',
                              'device-is-optical-disc' : False,
                            },
                        '/org/freedesktop/UDisks/devices/scd0' :
                            { 'device-is-optical-disc' : True,
                            },
                       }
        dbus.service.Object.__init__(self, bus, path)

        self.device_objects = []
        for d in self.devices:
            self.device_objects.append(FakeUDisksDevice(bus, d,
                                       self.devices[d]))

    @dbus.service.method(DISKS_IFACE, out_signature='ao')
    def EnumerateDevices(self):
        return list(self.devices.keys())

    @dbus.service.signal(DISKS_IFACE, signature='o')
    def DeviceAdded(self, path):
        pass

    @dbus.service.method(DISKS_IFACE)
    def emitDeviceAdded(self):
        self.DeviceAdded('/org/freedesktop/UDisks/devices/sdc')
    
    @dbus.service.signal(DISKS_IFACE, signature='o')
    def DeviceRemoved(self, path):
        pass

    @dbus.service.method(DISKS_IFACE)
    def emitDeviceRemoved(self):
        self.DeviceRemoved('/org/freedesktop/UDisks/devices/sdc')
    
class FakeUDisksDevice(dbus.service.Object):
    def __init__(self, bus, path, props):
        self.props = props
        self.bus_name = dbus.service.BusName(DISKS_IFACE,
                                             bus)
        dbus.service.Object.__init__(self, bus, path)

    @dbus.service.method(PROPS_IFACE, in_signature='ss', out_signature='v')
    def Get(self, iface, prop):
        return self.props[prop]

    @dbus.service.method(DEVICE_IFACE, in_signature='sas', out_signature='s')
    def FilesystemMount(self, filesystem_type, options):
        return '/tmp'
        
if __name__ == '__main__':
    import gobject
    f = FakeUDisks(dbus.SessionBus(), '/org/freedesktop/UDisks')
    gobject.MainLoop().run()



'''

altgraph - a python graph library
=================================

altgraph is a fork of graphlib U{http://pygraphlib.sourceforge.net} tailored
to use newer Python 2.3+ features, including additional support used by the
py2app suite (modulegraph and macholib, specifically).

altgraph is a python based graph (network) representation and manipulation package.
It has started out as an extension to the C{graph_lib} module written by Nathan Denny
U{http://www.ece.arizona.edu/~denny/python_nest/graph_lib_1.0.1.html} but over time
it has been significantly optimized and expanded.

The L{Graph} class is loosely modeled after the U{LEDA<http://www.algorithmic-solutions.com/enleda.htm>} (Library of Efficient Datatypes)  representation. The library
includes methods for constructing graphs, BFS and DFS traversals,
topological sort, finding connected components, shortest paths as well as a number
graph statistics functions. The library can also visualize graphs
via U{graphviz<http://www.research.att.com/sw/tools/graphviz/>}.

The package contains the following modules:
    -  the L{Graph} module contains the L{Graph.Graph} class that stores the graph data
    -  the L{GraphAlgo} implements graph algorithms operating on graphs (L{Graph.Graph} instances)
    -  the L{GraphStat} contains functions for computing statistical measures on graphs
    -  the L{GraphUtil} contains functions for generating, reading and saving graphs
    -  the L{Dot}  contains functions for displaying graphs via U{graphviz<http://www.research.att.com/sw/tools/graphviz/>}.

Download
--------

Download the  U{python source code<../dist>}.

Installation
------------

Download and unpack the archive then type::

    python setup.py install

This will install the library in the default location. For instructions on
how to customize the install procedure read the output of::

    python setup.py --help install

To verify the installation procedure switch to the test directory and run the
test_graphib.py file that is located there::

    python test_altgraph.py

Example usage
-------------

Lets assume that we want to analyze the graph below (links to the full picture) GRAPH_IMG.
Our script then might look the following way::

    from altgraph import Graph, GraphAlgo, Dot

    # these are the edges
    edges = [ (1,2), (2,4), (1,3), (2,4), (3,4), (4,5), (6,5),
    (6,14), (14,15), (6, 15),  (5,7), (7, 8), (7,13), (12,8),
    (8,13), (11,12), (11,9), (13,11), (9,13), (13,10) ]

    # creates the graph
    graph = Graph.Graph()
    for head, tail in edges:
        graph.add_edge(head, tail)

    # do a forward bfs from 1 at most to 20
    print graph.forw_bfs(1)

This will print the nodes in some breadth first order::

    [1, 2, 3, 4, 5, 7, 8, 13, 11, 10, 12, 9]

If we wanted to get the hop-distance from node 1 to node 8
we coud write::

    print graph.get_hops(1, 8)

This will print the following::

    [(1, 0), (2, 1), (3, 1), (4, 2), (5, 3), (7, 4), (8, 5)]

Node 1 is at 0 hops since it is the starting node, nodes 2,3 are 1 hop away ...
node 8 is 5 hops away. To find the shortest distance between two nodes you
can use::

    print GraphAlgo.shortest_path(graph, 1, 12)

It will print the nodes on one (if there are more) the shortest paths::

    [1, 2, 4, 5, 7, 13, 11, 12]

To display the graph we can use the GraphViz backend::

    dot = Dot.Dot(graph)

    # display the graph on the monitor
    dot.display()

    # save it in an image file
    dot.save_img(file_name='graph', file_type='gif')


@author: U{Istvan Albert<http://www.personal.psu.edu/staff/i/u/iua1/>}

@license:  MIT License

Copyright (c) 2004 Istvan Albert unless otherwise noted.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do
so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
@requires: Python 2.3 or higher

@newfield contributor: Contributors:
@contributor: U{Reka Albert <http://www.phys.psu.edu/~ralbert/>}

'''

__version__ = '0.6.7'

class GraphError(ValueError):
    pass



"""
Python 2.4-like compatibility library for Python 2.3
"""
from itertools import *

#
# builtins from 2.4
#

try:
    set, frozenset
except NameError:
    from sets import Set as set, ImmutableSet as frozenset

try:
    sorted
except NameError:
    def sorted(iterable, cmp=None, key=None, reverse=False):
        if key is not None:
            a, b = tee(iterable)
            iterable = izip(imap(key, iterable), iterable)
        if cmp is not None:
            iterable = list(iterable)
            iterable.sort(cmp)
        else:
            iterable = isorted(iterable)
        if key is not None:
            iterable = [v for (k,v) in iterable]
        if type(iterable) is not list:
            iterable = list(iterable)
        if reverse:
            iterable.reverse()
        return iterable

try:
    reversed
except NameError:
    def reversed(iterable):
        lst = list(iterable)
        pop = lst.pop
        while lst:
            yield pop()


#
# itertools functions from 2.4
#
try:
    tee
except NameError:
    def tee(iterable, n=2):
        def gen(next, data={}, cnt=[0]):
            for i in count():
                if i == cnt[0]:
                    item = data[i] = next()
                    cnt[0] += 1
                else:
                    item = data.pop(i)
                yield item
        return tuple(imap(gen, repeat(iter(iterable), n)))

try:
    groupby
except NameError:
    class groupby(object):
        def __init__(self, iterable, key=None):
            if key is None:
                key = lambda x: x
            self.keyfunc = key
            self.it = iter(iterable)
            self.tgtkey = self.currkey = self.currvalue = xrange(0)
        def __iter__(self):
            return self
        def next(self):
            while self.currkey == self.tgtkey:
                self.currvalue = self.it.next() # Exit on StopIteration
                self.currkey = self.keyfunc(self.currvalue)
            self.tgtkey = self.currkey
            return (self.currkey, self._grouper(self.tgtkey))
        def _grouper(self, tgtkey):
            while self.currkey == tgtkey:
                yield self.currvalue
                self.currvalue = self.it.next() # Exit on StopIteration
                self.currkey = self.keyfunc(self.currvalue)


#
# operators from 2.4
#
try:
    from operator import attrgetter, itemgetter
except ImportError:
    def attrgetter(attr):
        def attrgetter(obj):
            return getattr(obj, attr)
        return attrgetter

    def itemgetter(item):
        def itemgetter(obj):
            return obj[item]
        return itemgetter


#
# deque from 2.4's collections
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/259179/
#
try:
    from collections import deque
except ImportError:
    class deque(object):

        def __init__(self, iterable=()):
            self.data = dict(enumerate(iterable))
            self.left = 0
            self.right = len(self.data)

        def append(self, x):
            self.data[self.right] = x
            self.right += 1

        def appendleft(self, x):
            self.left -= 1
            self.data[self.left] = x

        def pop(self):
            if self.left == self.right:
                raise IndexError('cannot pop from empty deque')
            self.right -= 1
            return self.data[self.right]

        def popleft(self):
            if self.left == self.right:
                raise IndexError('cannot pop from empty deque')
            x = self.data[self.left]
            self.left += 1
            return x

        def __len__(self):
            return self.right - self.left

        def __iter__(self):
            return imap(self.data.__getitem__, xrange(self.left, self.right))

        def __repr__(self):
            return 'deque(%r)' % (list(self),)

        def __getstate__(self):
            return (tuple(self),)

        def __setstate__(self, s):
            self.__init__(s[0])

        def __hash__(self):
            raise TypeError

        def __copy__(self):
            return self.__class__(self)

        def __deepcopy__(self, memo={}):
            from copy import deepcopy
            result = self.__class__()
            memo[id(self)] = result
            result.__init__(deepcopy(tuple(self), memo))
            return result

#
# new functions
#
import heapq as _heapq
def isorted(iterable):
    lst = list(iterable)
    _heapq.heapify(lst)
    pop = _heapq.heappop
    while lst:
        yield pop(lst)

def ireversed(iterable):
    if isinstance(iterable, (list, tuple)):
        for i in xrange(len(iterable)-1, -1, -1):
            yield iterable[i]
    for obj in reversed(iterable):
        yield obj



#
# Copyright (c) 2007, 2008 Agostino Russo
#
# Written by Agostino Russo <agostino.russo@gmail.com>
#
# win32.ui is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# win32.ui is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

'''
Python wrappers around win32 widgets and window classes
'''

import sys
import os

from usbcreator.frontends.winui.winui.defs import *
from usbcreator.misc import callable, text_type

__all__ = ["Window", "Frontend"]

#TBD use weakref in _event_handlers_
_event_handlers_ = {}

if sys.version.startswith('2.3'):
    from sets import Set as set

def event_dispatcher(hwnd, message, wparam, lparam):
    eh = _event_handlers_
    na = {}
    nas = set()
    handlers = set()
    for eh in [eh.get(hwnd, na), eh.get(None, na)]:
        for eh in [eh.get(message, na) , eh.get(None, na)]:
            for eh in [eh.get(wparam, na) , eh.get(None, na)]:
                if message == WM_NOTIFY:
                    try:
                        nmh = cast(lparam, POINTER(NMHDR)).contents
                        handlers |= eh.get(nmh.hwndFrom, na)
                        handlers |= eh.get(None, nas)
                    except:
                        pass
                else:
                    handlers |= eh.get(lparam, nas)
                    handlers |= eh.get(None, nas)
    for obj, handler_name in handlers:
        handler = getattr(obj, handler_name, None)
        if not callable(handler): continue
        result = handler((hwnd, message, wparam, lparam))
        if bool(result):
            return result
    return windll.user32.DefWindowProcW(
        c_int(hwnd),
        c_int(message),
        c_int(wparam),
        c_int(lparam))

def event_handler(hwnd=None, message=None, wparam=None, lparam=None):
    '''
    Decorator for event handlers
    The handler must return True to stop further message processing by other handlers of the same Window
    '''
    def decorator(func):
        func._handled_event_ = (hwnd, message, wparam, lparam)
        return func
    return decorator

class BasicWindow(object):
    '''
    Wraps
    '''
    _window_class_name_ = None
    _window_class_style_ = CS_HREDRAW | CS_VREDRAW
    _window_style_ = 0
    _window_ex_style_ = 0

    def __init__(self, parent=None, x=None, y=None, width=None, height=None, text=None, frontend=None, icon=None):
        self.parent = parent
        self._icon = None
        if frontend:
            self.frontend = frontend
        else:
            self.frontend = parent.frontend
        if not self.__class__._window_class_name_:
            self.__class__._window_class_name_ = self.__class__.__name__
            if icon:
                self._icon = windll.user32.LoadImageW(NULL, text_type(icon), IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
            self._register_window()
        self._create_window(x, y, width, height, text)
        self._register_handlers()
        self.on_init()

    def _register_window(self):
        self._window_class_ = WNDCLASSEX(
                event_dispatcher,
                self._window_class_name_,
                self._window_class_style_,
                icon=self._icon)
        self._window_class_._atom_ = windll.user32.RegisterClassExW(byref(self._window_class_))
        if not self._window_class_._atom_:
            raise WinError()

    def _create_window(self, x=None, y=None, width=None, height=None, text=None):
        hmenu = NULL
        lpparam = NULL
        hwnd = self.parent and self.parent._hwnd or NULL
        frontend_hinstance = self.frontend._hinstance
        if x is None:
            x = CW_USEDEFAULT
        if y is None:
            y = CW_USEDEFAULT
        if width is None:
            width = CW_USEDEFAULT
        if height is None:
            height = CW_USEDEFAULT
        if not text:
            text = ""
        self._hwnd = CreateWindowEx(
            self._window_ex_style_,
            text_type(self._window_class_name_),
            text_type(text),
            self._window_style_,
            x, y, width, height,
            hwnd,
            hmenu,
            frontend_hinstance,
            lpparam)

    def _register_handlers(self):
        for key in dir(self):
            handler = getattr(self, key)
            handled_event = getattr(handler, "_handled_event_", None)
            if callable(handler) and handled_event:
                handled_event = list(handled_event)
                for i,subkey in enumerate(handled_event):
                    if subkey is SELF_HWND:
                        handled_event[i] = self._hwnd
                    elif subkey is PARENT_HWND:
                        handled_event[i] = self.parent._hwnd
                    elif subkey is APPLICATION_HINSTANCE:
                        handled_event[i] = self.frontend._hinstance
                self._add_event_handler(key , handled_event)

    def _add_event_handler(self, handler_name, handled_event):
        hwnd, message, wparam, lparam = handled_event
        eh = _event_handlers_
        eh = eh.setdefault(hwnd,{})
        eh = eh.setdefault(message,{})
        eh = eh.setdefault(wparam,{})
        handlers = eh.setdefault(lparam,set())
        handlers.add((self, handler_name))

    def on_init(self):
        pass

def open_dialog(hwnd, file_filter=None):
    fn = OPENFILENAME()
    if file_filter:
        fn.lpstrFilter = LPCTSTR(file_filter)
        fn.nFilterIndex = 1;
    fn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST
    fn.nMaxFile = MAX_PATH
    fn.hwndOwner = hwnd
    fn.lStructSize = sizeof(fn)
    buf = create_unicode_buffer(MAX_PATH)
    fn.lpstrFile = cast(buf, c_wchar_p)
    if GetOpenFileName(byref(fn)) != 0:
        return fn.lpstrFile
    else:
        return ''
    
class Window(BasicWindow):
    _repaint_on_move_ = False
    _is_transparent_ = False

    def __init__(self, parent=None, x=None, y=None, width=None, height=None, text=None, frontend=None, icon=None):
        self._gdi_disposables = []
        self._background_color = None
        self._background_brush = None
        self._default_background_brush = None
        self._text_color = None
        self._null_brush = windll.gdi32.GetStockObject(NULL_BRUSH)
        self._gdi_disposables.append(self._null_brush)
        BasicWindow.__init__(self, parent, x, y, width, height, text, frontend, icon)
        self.set_font()
        self.update()

    def get_window_rect(self):
        rect = RECT()
        windll.user32.GetWindowRect(self._hwnd, byref(rect))
        return rect.left, rect.top, rect.right, rect.bottom

    def get_client_rect(self):
        rect = RECT()
        windll.user32.GetClientRect(self._hwnd, byref(rect))
        return rect.left, rect.top, rect.right, rect.bottom

    def move(self, x, y):
        if not windll.user32.SetWindowPos(self._hwnd, NULL, x, y, 0, 0,  SWP_NOSIZE):
        #~ if not windll.user32.MoveWindow(self._hwnd, x, y, -1, -1, self._repaint_on_move_): #repaint
            raise WinError()

    def resize(self, width, height):
        if not windll.user32.SetWindowPos(self._hwnd, NULL, 0, 0, width, height,  SWP_NOMOVE):
            raise WinError()

    def show(self):
        #http://msdn.microsoft.com/en-us/library/ms633548.aspx
        self.on_show()
        if not windll.user32.ShowWindow(self._hwnd, SW_SHOWNORMAL):
            #raise WinError()
            pass
        self.update()

    def enable(self):
        if not windll.user32.EnableWindow(self._hwnd, True):
            #raise WinError()
            pass
        self.update()

    def disable(self):
        if not windll.user32.EnableWindow(self._hwnd, False):
            #raise WinError()
            pass
        self.update()

    def hide(self):
        if not windll.user32.ShowWindow(self._hwnd, SW_HIDE):
            #~ raise WinError()
            pass

    def set_focus(self):
        windll.user32.SetFocus(self._hwnd)

    def update(self, full=False):
        if full:
            windll.user32.ShowWindow(self._hwnd, SW_HIDE)
            windll.user32.ShowWindow(self._hwnd, SW_SHOW)
        if not windll.user32.UpdateWindow(self._hwnd):
            raise WinError()

    def get_text(self):
        buffer_max_len = 999
        buffer = (c_wchar * buffer_max_len)()
        if windll.user32.GetWindowTextW(self._hwnd, byref(buffer), buffer_max_len):
            return text_type(buffer.value)

    def set_text(self, text):
        old_text = self.get_text()
        if not windll.user32.SetWindowTextW(self._hwnd, text_type(text)):
            raise WinError()
        if old_text and old_text.rstrip() != text.rstrip():
            # without update, text is displayed on top of old text when background is transparent
            # TBD check _on_ctlcolorstatic whether that can be avoided
            self.update(full=True)

    def set_font(self, family='Tahoma', size=13, bold=False):
        weight = bold and FW_BOLD or FW_NORMAL
        font = windll.gdi32.CreateFontW(
            size, # height of font
            0, # average character width
            0, # angle of escapement
            0, # base-line orientation angle
            weight, # font weight
            0, # FALSE italic attribute option
            0, # FALSE underline attribute option
            0, # FALSE strikeout attribute option
            0, # DEFAULT_CHARSET character set identifier
            0, # OUT_DEFAULT_PRECIS output precision
            0, # CLIP_DEFAULT_PRECIS clipping precision
            0, # NONANTIALIASED_QUALITY output quality
            0, #0x20, DEFAULT_PITCH | FF_DONTCARE # pitch and family
            text_type(family) #TEXT("Verdana") # typeface name
            )
        self._gdi_disposables.append(font)
        self._send_message(WM_SETFONT, font, True)

    def set_background_color(self, red255=None, green255=None, blue255=None):
        if (red255, green255, blue255) == (None, None, None):
            self._background_color = None
            if self._default_background_brush:
                windll.user32.SetClassLongW(self._hwnd, GCL_HBRBACKGROUND, self._default_background_brush)
        else:
            self._background_color = RGB(red255, blue255, green255)
            self._background_brush = windll.gdi32.CreateSolidBrush(self._background_color)
            self._default_background_brush = windll.user32.SetClassLongW(self._hwnd, GCL_HBRBACKGROUND, self._background_brush)
            self._gdi_disposables.append(self._background_color)
            self._gdi_disposables.append(self._background_brush)
            self._gdi_disposables.append(self._default_background_brush)

    def set_text_color(self, red255=None, green255=None, blue255=None):
        if (red255, green255, blue255) == (None, None, None):
            self._text_color = None
        else:
            self._text_color = RGB(red255, green255, blue255)
            self._gdi_disposables.append(self._text_color)

    def set_transparency(self, transparent):
        self.is_transparent = transparent
        self.update(full = True)

    def _on_command(self, event):
        self.on_command(event)
    _on_command = event_handler(message=WM_COMMAND, lparam=SELF_HWND)(_on_command)

    def on_command(self, event):
        pass
    
    def _on_notify(self, event):
        self.on_notify(event)
    _on_notify = event_handler(message=WM_NOTIFY, lparam=SELF_HWND)(_on_notify)

    def on_notify(self, event):
        pass
    
    def _on_wm_hscroll(self, event):
        self.on_wm_hscroll(event)
    _on_wm_hscroll = event_handler(message=WM_HSCROLL, lparam=SELF_HWND)(_on_wm_hscroll)

    def on_wm_hscroll(self, event):
        pass

    def stop_redraw(self):
        self._send_message(WM_SETREDRAW, False, 0)
        #~ LockWindowUpdate(self._hwnd)
        #~ LockWindowUpdate(NULL)

    def start_redraw(self):
        self._send_message(WM_SETREDRAW, True, 0)

    def _send_message(self, message, wparam=0, lparam=0):
        return windll.user32.SendMessageW(self._hwnd, message, wparam, lparam)

    def _on_destroy(self, event):
        for x in self._gdi_disposables:
            try:
                windll.gdi32.DeleteObject(x)
            except:
                pass
        self.on_destroy()
    _on_destroy = event_handler(message=WM_DESTROY, hwnd=SELF_HWND)(_on_destroy)

    def _on_ctlcolorstatic(self, event):
        hdc = event[2]
        if self._text_color:
            windll.gdi32.SetTextColor(hdc, self._text_color)
        if self._is_transparent_:
            windll.gdi32.SetBkMode(hdc, TRANSPARENT)
            brush = self._null_brush
        else:
            brush = True
        return brush
    event_handler(message=WM_CTLCOLORBTN, lparam=SELF_HWND)(_on_ctlcolorstatic)
    event_handler(message=WM_CTLCOLORSTATIC, lparam=SELF_HWND)(_on_ctlcolorstatic)

    def on_show(self):
        pass

    def on_destroy(self):
        pass

class Frontend(object):
    '''
    Wraps a Windows application
    It is associated to a main window
    It controls the message processing
    '''
    _main_window_class_ = Window

    def __init__(self, main_window_class=None, **kargs):
        self._hwnd = None
        self._hinstance = windll.kernel32.GetModuleHandleW(c_int(NULL))
        kargs["frontend"] = self
        if not main_window_class:
            main_window_class = self._main_window_class_
        self.main_window = main_window_class(**kargs)
        self.on_init()

    def set_title(self, title):
        self.main_window.set_text(title)

    def set_icon(self, icon_path):
        if icon_path and os.path.isfile(icon_path):
            self.main_window._icon = windll.user32.LoadImageW(NULL, text_type(icon_path), IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
            windll.user32.SendMessageW(self.main_window._hwnd, WM_SETICON, ICON_SMALL, self.main_window._icon)

    def get_title(self):
        return self.main_window.get_text()

    def run(self):
        '''
        Starts the message processing
        '''
        msg = MSG()
        pMsg = pointer(msg)
        self._keep_running = True
        self.on_run()
        while self._keep_running and windll.user32.GetMessageW(pMsg, NULL, 0, 0) > 0:
            #TBD if IsDialogMessage is used, other messages are not processed, for now doing a manual exception
            if self.main_window._hwnd == NULL \
            or pMsg.contents.message in (WM_COMMAND, WM_PAINT, WM_CTLCOLORSTATIC, WM_DESTROY, WM_QUIT) \
            or not windll.user32.IsDialogMessage(self.main_window._hwnd , pMsg):
                windll.user32.TranslateMessage(pMsg)
                windll.user32.DispatchMessageW(pMsg)

    def stop(self):
        '''
        Stops the message processing
        '''
        self._keep_running = False
        #Post a message to unblock GetMessageW
        windll.user32.PostMessageW(self.main_window._hwnd,WM_NULL, 0, 0)

    def on_init(self):
        pass

    def on_run(self):
        pass

    def quit(self):
        '''
        Destroys the main window
        '''
        windll.user32.DestroyWindow(self.main_window._hwnd)

    def _quit(self):
        '''
        Really quit anything on the windows side,
        this is called by MainWindow.on_destroy
        '''
        windll.user32.PostQuitMessage(0)
        self.on_quit()

    def on_quit(self):
        pass

    def show_error_message(self, message, title=None):
        if not title:
            title = self.get_title()
        windll.user32.MessageBoxW(self.main_window._hwnd, text_type(message), text_type(title), MB_OK|MB_ICONERROR)

    def show_info_message(self, message, title=None):
        if not title:
            title = self.get_title()
        windll.user32.MessageBoxW(self.main_window._hwnd, text_type(message), text_type(title), MB_OK|MB_ICONINFORMATION)

    def ask_confirmation(self, message, title=None):
        if not title:
            title = self.get_title()
        result = windll.user32.MessageBoxW(self.main_window._hwnd, text_type(message), text_type(title), MB_YESNO|MB_ICONQUESTION)
        return result == IDYES

    def ask_to_retry(self, message, title=None):
        if not title:
            title = self.get_title()
        result = windll.user32.MessageBoxW(self.main_window._hwnd, text_type(message), text_type(title), MB_RETRYCANCEL)
        return result == IDRETRY

class MainWindow(Window):
    '''
    Main Window
    Has borders, and is overlapped, when it is closed, the frontend quits
    '''

    _window_class_name_ = None
    _window_style_ = WS_OVERLAPPEDWINDOW
    _window_ex_style_ = WS_EX_CONTROLPARENT

    def on_destroy(self):
        self.frontend._quit()

    def __del__(self):
        self.frontend._quit()

class MainDialogWindow(MainWindow):
    '''
    Like MainWindow
    But cannot be resized, looks like a dialog
    '''
    _window_style_ =  WS_BORDER |  WS_SYSMENU | WS_CAPTION

class Widget(Window):
    _window_class_name_ = "Must Override This"
    _window_style_ = WS_CHILD | WS_VISIBLE | WS_TABSTOP
    _window_ex_style_ = 0

class StaticWidget(Widget):
    _window_class_name_ = "STATIC"
    _window_style_ = WS_CHILD | WS_VISIBLE
    _window_ex_style_ = WS_EX_TRANSPARENT
    _is_transparent_ = True

class EtchedRectangle(Widget):
    _window_class_name_ = "STATIC"
    _window_style_ = WS_CHILD | WS_VISIBLE | SS_ETCHEDFRAME

class Panel(Window):
    _window_class_name_ = None
    _window_style_ = WS_CHILD | WS_VISIBLE
    _window_ex_style_ = WS_EX_CONTROLPARENT
    #_window_style_ = StaticWidget._window_style_|WS_BORDER

class Edit(Widget):
    _window_class_name_ = "EDIT"
    _window_style_ = Widget._window_style_ | WS_TABSTOP
    _window_ex_style_ = WS_EX_CLIENTEDGE

class PasswordEdit(Widget):
    _window_class_name_ = "EDIT"
    _window_style_ = Widget._window_style_ | ES_PASSWORD | WS_TABSTOP
    _window_ex_style_ = WS_EX_CLIENTEDGE

class Tab(Widget):
    #define WC_TABCONTROLW          L"SysTabControl32"   ???
    _window_class_name_ = "SysTabControl32"

    def add_item(self, title, child, position=0):
        item = TCITEM()
        item.mask = TCIF_TEXT | TCIF_PARAM
        item.pszText = text_type(title)
        item.lParam = child._hwnd
        #~ self.InsertItem(index, item)
        #~ self._ResizeChild(child)
        #~ self.SetCurrentTab(index)
        self._send_message(TCM_INSERTITEM, position, byref(item))

class Tooltip(Widget):
    _window_class_name_ = u"SysTabControl32"

class Trackbar(Widget):
    _window_class_name_ = u'msctls_trackbar32'

    def set_range(self, minimum, maximum):
        #r = ((minimum & 0xFFFF) | (maximum & 0xFFFF)) << 16
        #self._send_message(TBM_SETRANGE, TRUE, r)
        self._send_message(TBM_SETRANGEMIN, TRUE, minimum)
        self._send_message(TBM_SETRANGEMAX, TRUE, maximum)

    def get_pos(self):
        return self._send_message(TBM_GETPOS, 0, 0)

    def set_buddy(self, hwnd, left=True):
        if left:
            left = TRUE
        else:
            left = FALSE
        return self._send_message(TBM_SETBUDDY, left, hwnd)

    def on_wm_hscroll(self, event):
        pass

class ListView(Widget):
    _window_class_name_ = u'SysListView32'
    _window_style_ = Widget._window_style_ | WS_BORDER | LVS_REPORT | LVS_SINGLESEL

    def on_notify(self, event):
        lparam = event[3]
        nmh = cast(lparam, POINTER(NMLISTVIEW)).contents
        if nmh.hdr.code == LVN_ITEMCHANGED:
            if (nmh.uNewState & LVIS_SELECTED) != 0:
                self.on_selection_changed(nmh.iItem)

    def on_selection_changed(self, item):
        pass

    def clear(self):
        self._send_message(LVM_DELETEALLITEMS, 0, 0)

    def insert_column(self, col_name, col_width, idx=0):
        c = LVCOLUMN()
        c.mask = LVCF_TEXT | LVCF_WIDTH | LVCF_SUBITEM
        c.pszText = text_type(col_name)
        c.cx = col_width
        self._send_message(LVM_INSERTCOLUMN, idx, byref(c))
    
    def insert_item(self, item_name, idx):
        if idx < 0:
            # Append.
            idx = self._send_message(LVM_GETITEMCOUNT, 0, 0)
        item = LVITEM()
        item.mask = LVIF_TEXT
        item.iItem = idx
        item.iSubItem = 0
        item.pszText = text_type(item_name)
        return self._send_message(LVM_INSERTITEM, 0, byref(item))

    def delete_item(self, idx):
        self._send_message(LVM_DELETEITEM, idx, 0)

    def set_item(self, idx, subidx, text):
        item = LVITEM()
        item.mask = LVIF_TEXT
        item.iItem = idx
        item.iSubItem = subidx
        item.pszText = text_type(text)
        self._send_message(LVM_SETITEM, 0, byref(item))

    def get_selection(self, sub_index=None):
        idx = self._send_message(LVM_GETNEXTITEM, -1, LVNI_SELECTED)
        if idx == -1:
            return ''
        item = LVITEM()
        item.iItem = idx
        item.mask = LVIF_TEXT
        if sub_index != None:
            item.iSubItem = sub_index
        # XXX evand 2009-07-21: Using MAX_PATH here because the UDI for ISO
        # files will be their filename.
        item.cchTextMax = MAX_PATH
        buf = create_unicode_buffer(MAX_PATH)
        item.pszText = cast(buf, c_wchar_p)
        self._send_message(LVM_GETITEM, 0, byref(item))
        return item.pszText
        

class ListBox(Widget):
    _window_class_name_ = "LISTBOX"
    _window_style_ = Widget._window_style_ | LBS_NOTIFY | WS_BORDER | WS_VSCROLL | WS_TABSTOP

    def add_item(self, text):
        self._send_message(LB_ADDSTRING, 0, text_type(text))

class SortedListBox(Widget):
    _window_style_ = ListBox._window_style_ | LBS_SORT

class ComboBox(Widget):
    _window_class_name_ = "COMBOBOX" #"ComboBoxEx32"
    _window_style_ = Widget._window_style_ | CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP

    def on_command(self, event):
        if event[2] == 589824: #TBD use a constant name
            self.on_change()

    def set_value(self, value):
        self._send_message(CB_SELECTSTRING, -1, text_type(value)) # CB_SETCURSEL, value, 0)

    def add_item(self, text):
        self._send_message(CB_ADDSTRING, 0, text_type(text))

    def on_change(self):
        pass

class SortedComboBox(ComboBox):
    _window_style_ = ComboBox._window_style_ | CBS_SORT

class Button(Widget):
    _window_class_name_ = "BUTTON"
    _window_style_ = Widget._window_style_ | BS_PUSHBUTTON | WS_TABSTOP
    #~ _is_transparent_ = True
    #~ _window_ex_style_ = WS_EX_TRANSPARENT

    def on_command(self, event):
        if event[2] == 0:
            self.on_click()

    def is_checked(self):
        value = self._send_message(BM_GETCHECK, 0, 0)
        return value == BST_CHECKED

    def set_check(self, value):
        if value:
            value = BST_CHECKED
        else:
            value = BST_UNCHECKED
        self._send_message(BM_SETCHECK, value, 0)

    def on_click(self):
        pass

class FlatButton(Button):
    _window_style_ = Widget._window_style_ | BS_FLAT | WS_TABSTOP

class DefaultButton(Button):
    _window_style_ = Widget._window_style_ | BS_DEFPUSHBUTTON  | WS_TABSTOP

class RadioButton(Button):
    _window_class_name_ = "BUTTON"
    _window_style_ = Widget._window_style_ | BS_AUTORADIOBUTTON | WS_TABSTOP
    _is_transparent_ = True

class GroupBox(Button):
    _window_class_name_ = "BUTTON"
    _window_style_ = Widget._window_style_ | BS_GROUPBOX  #| WS_TABSTOP
    _is_transparent_ = True
    _window_ex_style_ = WS_EX_CONTROLPARENT

class CheckButton(Button):
    _window_class_name_ = "BUTTON"
    _window_style_ = Widget._window_style_ | BS_AUTOCHECKBOX  | WS_TABSTOP
    _is_transparent_ = True

class Label(StaticWidget):
    _window_class_name_ = "Static"
    _window_style_ = StaticWidget._window_style_ | SS_NOPREFIX

class Bitmap(StaticWidget):
    _window_class_name_ = "Static"
    _window_style_ = StaticWidget._window_style_ | SS_BITMAP
    _window_ex_style_ = WS_EX_TRANSPARENT

    def set_image(self, path, width=0, height=0):
        path = text_type(path)
        himage = windll.user32.LoadImageW(NULL, path, IMAGE_BITMAP, width, height, LR_LOADFROMFILE);
        self._gdi_disposables.append(himage)
        self._send_message(STM_SETIMAGE, IMAGE_BITMAP, himage)

class Icon(StaticWidget):
    _window_class_name_ = "Static"
    _window_style_ = StaticWidget._window_style_|SS_ICON

    def set_image(self, path, width=0, height=0):
        path = text_type(path)
        himage = windll.user32.LoadImageW(NULL, path, IMAGE_ICON, width, height, LR_LOADFROMFILE);
        self._gdi_disposables.append(himage)
        self._send_message(STM_SETIMAGE, IMAGE_ICON, himage)

class ProgressBar(Widget):
    _window_class_name_ = "msctls_progress32"

    def set_position(self, position):
        return self._send_message(PBM_SETPOS, position, 0)

    def get_position(self):
        return self._send_message(PBM_GETPOS, 0, 0)

    def set_bar_color(self, color):
        return self._send_message(PBM_SETBARCOLOR, 0, color)

    def set_background_color(self, clrBk):
        return self._send_message(PBM_SETBKCOLOR, 0, color)

    def step(self, n=1):
        step0 = self.get_position()
        #~ if not n:
            #~ step1 = self._send_message(PBM_STEPIT, 0, 0) #PBM_DELTAPOS
        step1 = self.set_position(step0 + n)
        self.on_change()
        return step1

    def on_change(self):
        pass

class Link(Widget):
    _window_class_name_ = "Link"

class Page(Window):
    _window_class_name_ = None
    _window_style_ = WS_CHILD
    _window_ex_style_ = WS_EX_CONTROLPARENT



# Copyright (C) 2009 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from usbcreator.frontends.base import Frontend
from usbcreator.frontends.winui.winui import ui
from usbcreator.frontends.winui.main_dialog import MainDialog
from usbcreator.frontends.winui.install_window import InstallWindow
from usbcreator import misc

class WinuiFrontend(Frontend, ui.Frontend):
    _main_window_class_ = ui.MainWindow

    def __init__(self, backend):
        #Frontend.__init__(self)
        ui.Frontend.__init__(self)
        self.current_page = None

        self.maindialog = MainDialog(self.main_window)
        self.installwindow = InstallWindow(self.main_window)

        # Connect signals.
        self.maindialog.quit_button.on_click = self.quit
        self.maindialog.sources.on_selection_changed = \
            self.selection_changed_source
        self.maindialog.targets.on_selection_changed = \
            self.selection_changed_target
        self.maindialog.no_persist_radio.on_click = \
            self.maindialog.persist_slider.disable
        self.maindialog.persist_radio.on_click = \
            self.maindialog.persist_slider.enable
        self.maindialog.persist_slider.on_wm_hscroll = \
            self.persist_value_changed
        self.maindialog.make_button.on_click = \
            self.install
        self.maindialog.other_button.on_click = \
            self.other
        # TODO evand 2009-07-24: This really needs a confirmation MessageBox.
        self.installwindow.cancel_button.on_click = \
            self.quit

        # Set initial state.
        self.maindialog.make_button.disable()

        # Connect to backend signals.
        self.backend = backend
        self.backend.source_added_cb = self.add_source
        self.backend.target_added_cb = self.add_target
        self.backend.failure_cb = self.failure
        self.backend.success_cb = self.success
        self.backend.install_progress_cb = self.progress
        self.backend.install_progress_message_cb = self.progress_message
        self.backend.retry_cb = self.retry

        # Scan for devices and bring up the window.
        self.backend.detect_devices()
        self.show_page(self.maindialog)

    def success(self):
        self.installwindow.hide()
        self.show_info_message(_('The installation is complete.  You may now '
                                 'reboot your computer with this device '
                                 'inserted to try or install Ubuntu.'),
                               _('Installation complete'))
        ui.Frontend.stop(self)

    def failure(self, message=None):
        logging.critical('Installation failed.')
        self.installwindow.hide()
        fail_title = _('Installation failed')
        if message:
            self.show_error_message(message, fail_title)
        else:
            # TODO evand 2009-07-28: Better error message.
            self.show_error_message(_('Installation failed.'), fail_title)
        ui.Frontend.stop(self)

    def retry(self, message):
        return self.ask_to_retry(message)

    # Frontend functions.
    def add_source(self, source):
        # FIXME evand 2009-07-21: Don't expose backend data structures like
        # this.
        source = self.backend.sources[source]
        i = self.maindialog.sources.insert_item(source['device'], -1)
        self.maindialog.sources.set_item(i, 1, source['label'])
        self.maindialog.sources.set_item(
            i, 2, misc.format_size(source['size']))

    def add_target(self, target):
        t = self.maindialog.targets
        t.clear()
        target = self.backend.targets[target]
        i = t.insert_item(target['device'], -1)
        t.set_item(i, 1, target['label'])
        t.set_item(i, 2, misc.format_size(target['capacity']))
        t.set_item(i, 3, misc.format_size(target['free']))

    # Callback functions.
    def selection_changed_target(self, item):
        sel = self.maindialog.targets.get_selection()
        target = self.backend.targets[sel]
        maximum = target['free'] / 1024 / 1024
        if maximum < misc.MIN_PERSISTENCE:
            # disable persistence.
            self.maindialog.persist_slider.disable()
        else:
            self.maindialog.persist_slider.enable()
            self.maindialog.persist_slider.set_range(
                misc.MIN_PERSISTENCE, maximum)
        self.selection_changed(item)

    def selection_changed_source(self, item):
        self.backend.set_current_source(self.maindialog.sources.get_selection())
        self.selection_changed(item)

    def selection_changed(self, item):
        # TODO evand 2009-07-29: Doesn't belong here.  Should go in the
        # target_changed callback.
        s = self.maindialog.sources
        t = self.maindialog.targets
        if s.get_selection() and t.get_selection():
            self.maindialog.make_button.enable()
        else:
            self.maindialog.make_button.disable()

    def persist_value_changed(self, event):
        pos = self.maindialog.persist_slider.get_pos()
        self.maindialog.persist_label.set_text('%d MB' % pos)

    def quit(self):
        self.backend.cancel_install()
        ui.Frontend.quit(self)

    def progress(self, complete, remaining, speed):
        self.installwindow.progress_bar.set_position(complete)
        if remaining and speed:
            # TODO evand 2009-07-24: Could use a time formatting function like
            # our human size function.
            mins = int(remaining / 60)
            secs = int(remaining % 60)
            text = _('%d%% complete (%dm%ss remaining)') % \
                    (complete, mins, secs)
            self.installwindow.progress_subtitle.set_text(text)
        else:
            self.installwindow.progress_subtitle.set_text(_('%d%% complete') %
                                                          complete)

    def progress_message(self, message):
        self.installwindow.progress_title.set_text(message)

    def install(self):
        source = misc.text_type(self.maindialog.sources.get_selection())
        target = misc.text_type(self.maindialog.targets.get_selection())
        if self.maindialog.persist_radio.is_checked():
            persist = self.maindialog.persist_slider.get_pos()
        else:
            persist = 0
        if source and target:
            self.installwindow.show()
            self.main_window.hide()
            self.backend.install(source, target, persist)

    def other(self):
        # TODO evand 2009-07-28: The list itself needs to be moved into the
        # base frontend.
        # To be displayed as a list of file type filters.
        filters = [ _('CD Images'), '*.ISO',
                    _('Disk Images'), '*.IMG',
                    _('All'), '*.*'
                  ]
        f = '\0'.join(filters) + '\0'
        filename = ui.open_dialog(self.main_window._hwnd, file_filter=f)
        if filename != '':
            self.backend.add_image(filename)

    # Everything else.

    def on_init(self):
        self.main_window.set_text(_('Make Startup Disk'))
        self.main_window.resize(375,425)

    def show_page(self, page):
        if self.current_page is page:
            self.current_page.show()
            return
        if self.current_page:
            self.current_page.hide()
        self.current_page = page
        page.show()
        self.main_window.show()
        self.run()



from distutils.core import setup
from DistUtilsExtra.command import *
import os

class usb_creator_build_i18n(build_i18n.build_i18n):
    def run(self):
         build_i18n.build_i18n.run(self)
         print("extracting strings for KDE frontend")
         os.system("./Messages.sh")

setup(name='usb-creator',
    version='0.2.23',
    description='Ubuntu startup disk creator',
    author='Evan Dandrea',
    author_email='evand@ubuntu.com',
    packages=['usbcreator',
              'usbcreator.frontends',
              'usbcreator.frontends.gtk',
              'usbcreator.frontends.kde',
              'usbcreator.frontends.base',
              'usbcreator.backends',
              'usbcreator.backends.base',
              'usbcreator.backends.fastboot',
              'usbcreator.backends.udisks',
             ],
    scripts=['bin/usb-creator-gtk','bin/usb-creator-kde'],
    data_files=[('share/usb-creator', ['gui/usbcreator-gtk.ui']),
                ('share/usb-creator', ['bin/usb-creator-helper']),
                ('share/usb-creator', ['gui/ubuntu-nexus7-USAGE-NOTICE-en.txt']),
                ('share/icons/hicolor/scalable/apps', ['desktop/usb-creator-gtk.svg', 'desktop/usb-creator-kde.svg']),
                ('share/kde4/apps/usb-creator-kde', ['gui/usbcreator-kde.ui']),
                ('/etc/dbus-1/system.d', ['dbus/com.ubuntu.USBCreator.conf']),
                ('share/dbus-1/system-services', ['dbus/com.ubuntu.USBCreator.service']),
                ('share/apport/package-hooks', ['debian/source_usb-creator.py'])],
    cmdclass = { "build" : build_extra.build_extra,
        "build_i18n" :  usb_creator_build_i18n,
        "build_help" :  build_help.build_help,
        "build_icons" :  build_icons.build_icons,
        "clean": clean_i18n.clean_i18n,
        }
    )




import logging

from gi.repository import GUdev
from gi.repository import GLib

from usbcreator.backends.base import Backend
from usbcreator import misc

KNOWN_IDS = {
    'ID_VENDOR_ID': ('18d1',),
    'ID_MODEL_ID': ('4e40', 'd001',),
}

class FastbootBackend(Backend):
    def __init__(self):
        Backend.__init__(self)
        logging.debug('FastbootBackend')
        self.client = GUdev.Client(subsystems=['usb'])
        
    def on_uevent(self, action, device):
        logging.debug('action: %s' % action)
        logging.debug('device: %s' % device.get_sysfs_path())

        for key,ids in KNOWN_IDS.items():
            result = device.get_property(key)
            if result not in ids:
                logging.debug('Unknown %s: %s' % (key, result))
                return
            
        [logging.debug('%s=%s' % (k, device.get_property(k))) for k in device.get_property_keys()]
        
        key = device.get_property('ID_SERIAL_SHORT')
        
        if action == 'add':
            self.targets[key] = {
                'vendor'     : device.get_property('ID_VENDOR_FROM_DATABASE'),
                'model'      : device.get_property('ID_MODEL'),
                'label'      : '',
                'device'     : device.get_property('ID_SERIAL_SHORT'),
                'status'     : misc.CAN_USE,
                }
            if misc.callable(self.target_added_cb):
                self.target_added_cb(key)
        elif action == 'remove':
            self._device_removed(key)
            
    def detect_devices(self):
        def _on_uevent(client, action, device):
            self.on_uevent(action, device)
            
        self.client.connect('uevent', _on_uevent)
        # Just in case, go over already attached devices
        for device in self.client.query_by_subsystem('usb'):
            self.on_uevent('add', device)

    def update_free(self):
        # No progress yet.
        return True

    def _is_casper_cd(self, filename):
        # no cds for us
        return None    

    def unmount(self):
        return

    def shutdown(self):
        return



# Copyright (C) 2009 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from usbcreator.frontends.winui.winui import ui

class MainDialog(ui.Page):
    '''Main window of usb-creator.'''
    def on_init(self):
        self.heading = ui.Label(self.parent, 10, 4, 365, 25,
            _('To try or install Ubuntu from a removable disk, '
              'it needs to be set up as a startup disk.'))
        self.source_heading = ui.Label(self.parent, 10, 30, 365, 25,
                                       _('Source disc image (.iso) or CD:'))
        listview_width = 350
        self.sources = ui.ListView(self.parent, 10, 50, listview_width, 100)
        self.sources.insert_column(_('Image'), listview_width / 3, idx=0)
        self.sources.insert_column(_('OS Version'), listview_width / 3, idx=1)
        self.sources.insert_column(_('Size'), listview_width / 3, idx=2)
        
        self.other_button = ui.Button(self.parent, 310, 155, 50, 25,
                                      _('Other...'))
        self.target_heading = ui.Label(self.parent, 10, 170, 365, 25,
                                       _('Removable disk to use:'))
        self.targets = ui.ListView(self.parent, 10, 185, listview_width, 100)
        self.targets.insert_column(_('Device'), listview_width / 4, idx=0)
        self.targets.insert_column(_('Label'), listview_width / 4, idx=1)
        self.targets.insert_column(_('Capacity'), listview_width / 4, idx=2)
        self.targets.insert_column(_('Free Space'), listview_width / 4, idx=3)
        
        self.persistent_heading = ui.Label(self.parent, 10, 285, 365, 25,
                                           _('When starting up from this disk, '
                                             'documents and settings will be:'))
        self.persist_radio = ui.RadioButton(self.parent, 10, 295, 365, 25,
                                            _('Stored in reserved space'))
        self.how_much = ui.Label(self.parent, 30, 320, 75, 25, _('How much:'))
        self.persist_label = ui.Label(self.parent, 0, 0, 50, 25, '')
        self.persist_slider = ui.Trackbar(self.parent, 90, 315, 225, 25)
        self.persist_slider.set_buddy(self.persist_label._hwnd, left=False)
        self.persist_radio.set_check(True)
        self.no_persist_radio = ui.RadioButton(self.parent, 10, 335, 365, 25,
                                               _('Discarded on shutdown, unless'
                                                 ' you save them elsewhere'))

        self.quit_button = ui.DefaultButton(self.parent, 200, 365, 50, 25,
                                            _('Quit'))
        self.make_button = ui.Button(self.parent, 255, 365, 100, 25,
                                     _('Make startup disk'))
